VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "CExpressionEvaluator"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Attribute VB_Ext_KEY = "SavedWithClassBuilder" ,"No"
Attribute VB_Ext_KEY = "Top_Level" ,"Yes"

' Class       : CExpressionEvaluator
' Description : This class implements expression evaluation.
' Source      : Total VB SourceBook 6

'Recognized tokens
Private Const mcstrMult As String = "*"
Private Const mcstrDiv As String = "/"
Private Const mcstrMod As String = "%"
Private Const mcstrPower As String = "^"
Private Const mcstrAdd As String = "+"
Private Const mcstrSub As String = "-"
Private Const mcstrLParen As String = "("
Private Const mcstrRParen As String = ")"
Private Const mcstrUnaryMinus As String = "-"
Private Const mcstrUnaryPlus As String = "+"
Private Const mcstrSin As String = "sin"
Private Const mcstrCos As String = "cos"
Private Const mcstrTan As String = "tan"
Private Const mcstrAtan As String = "atn"
Private Const mcstrExp As String = "exp"
Private Const mcstrLog As String = "log"
Private Const mcstrPi As String = "pi"
Private Const mcstrE As String = "e"

' Constant values
Private Const mcdblPi As Double = 3.14159265358979
Private Const mcdblE As Double = 2.718282

' variables for static property data
Private m_strExpression As String
Private m_strPostfix As String

' Expression variables
Private mfNegateParenOps As Boolean
Private mstrToken As String
Private mstrSeparatorUsed As String

' Token variables
Private Enum tokTokenState
  tokWhiteSpace
  tokToken
  tokUndefined
End Enum

Private Enum tokCharType
  tokWhiteSpaceChar
  tokSeparatorChar
  tokTokenChar
End Enum

Private mstrWhiteSpaceChars As String
Private mstrSeparatorChars As String
Private meState As tokTokenState
Private mstrCurrentChar As String
Private mstrInternalToken As String
Private mlngLen As Long
Private mlngPos As Long

Private mstrTokenText As String

' Stack variables
' Internal array variable to store stack items
Private mavarData() As Variant
' The index of the top item in the stack
Private mintTop As Integer
' The size of the stack. The stack is used in evaluating expressions
' If this class runs out of stack space this constant can be increased
Private Const mcintSizeStack As Integer = 200

Private Sub Class_Initialize()
  ' Set initial values to defaults which may be overridden
  ' with property settings
  ' Source: Total VB SourceBook 6
  '
  On Error GoTo PROC_ERR

  ' initialize stack
  ReDim mavarData(mcintSizeStack)
  
  'set the index into the array to zero
  mintTop = 0
  
PROC_EXIT:
  Exit Sub

PROC_ERR:
  MsgBox "Error: " & Err.Number & ". " & Err.Description, , _
    "Class_Initialize"
  Resume PROC_EXIT
  
End Sub

Public Property Get Expression() As String
  ' Returns: the expression string
  ' Source: Total VB SourceBook 6

  Expression = m_strExpression

End Property

Public Property Let Expression(ByVal strValue As String)
  ' strValue: Set the expression string
  ' Source: Total VB SourceBook 6

  m_strExpression = strValue
  
  ' Create the expression to postfix
  ConvertToPostfix

End Property

Public Property Get Postfix() As String
  ' Returns: the postfix representation of the expression string
  ' Source: Total VB SourceBook 6

  Postfix = m_strPostfix

End Property

Public Function Evaluate() As Double
  ' Comments  : This function evaluates the expression. For simplicity the
  '             postfix representation of the expression is evaluated
  ' Parameters: None
  ' Returns   : The result of the expression
  ' Source    : Total VB SourceBook 6
  '
  Dim dblNumber1 As Double
  Dim dblNumber2 As Double
  Dim dblResult As Double
  Dim strToken As String
  Dim fUnaryMinus As Boolean
  
  On Error GoTo PROC_ERR
  
  ' Reset the characters used as separators
  mstrSeparatorChars = vbNullString
  
  ' Set the string used by the tokenizer
  SetTokenString m_strPostfix
  
  ' while there are tokens in the expression
  Do While GetNextToken(strToken, mstrSeparatorUsed)
    ' if this token is an operator
    If strToken = mcstrMult Or strToken = mcstrAdd Or strToken = mcstrDiv Or _
      strToken = mcstrSub Or strToken = mcstrPower Or strToken = mcstrMod Then
      
      ' Get the top two values on the stack
      PopResult dblNumber2
      PopResult dblNumber1
      
      ' Perform the operation indicated by strToken
      Select Case strToken
        Case mcstrAdd
          dblResult = dblNumber1 + dblNumber2
        Case mcstrMult
          dblResult = dblNumber1 * dblNumber2
        Case mcstrDiv
          dblResult = dblNumber1 / dblNumber2
        Case mcstrSub
          dblResult = dblNumber1 - dblNumber2
        Case mcstrPower
          dblResult = dblNumber1 ^ dblNumber2
        Case mcstrMod
          dblResult = dblNumber1 Mod dblNumber2
      End Select
      
      ' store the result on the top of the stack
      PushResult dblResult
      
    Else
      ' This token is an operand
      
      ' Check for unary minus
      If Mid$(strToken, 1, 1) = mcstrUnaryMinus Then
        fUnaryMinus = True
        strToken = Right$(strToken, Len(strToken) - 1)
      Else
        fUnaryMinus = False
      End If
      
      ' If the token is a function
      If strToken = mcstrSin Or strToken = mcstrCos Or strToken = mcstrTan Or _
        strToken = mcstrAtan Or strToken = mcstrExp Or strToken = mcstrLog Then
        
        ' Get the topmost value on the stack. This is used by the function
        PopResult dblNumber1
        
        ' Perform the operation indicated by strToken
        Select Case strToken
          Case mcstrSin
            dblResult = Sin(dblNumber1)
          Case mcstrCos
            dblResult = Cos(dblNumber1)
          Case mcstrTan
            dblResult = Tan(dblNumber1)
          Case mcstrAtan
            dblResult = Atn(dblNumber1)
          Case mcstrExp
            dblResult = Exp(dblNumber1)
          Case mcstrLog
            dblResult = Log(dblNumber1)
        End Select
        
        If fUnaryMinus Then
          dblResult = -dblResult
        End If
        
        ' store the result on the top of the stack
        PushResult dblResult
      
      ElseIf strToken = mcstrPi Or strToken = mcstrE Then
        ' The token is a constant
        
        Select Case strToken
          Case mcstrPi
            dblResult = mcdblPi
          Case mcstrE
            dblResult = mcdblE
        End Select
        
        ' store the result on the top of the stack
        PushResult dblResult
      
      Else
        ' the token is a number
        dblResult = strToken
      
        If fUnaryMinus Then
          dblResult = -dblResult
        End If
        
        ' store the result on the top of the stack
        PushResult dblResult

      End If
    End If
  Loop
  
  ' The result of the evaluation is the topmost item on the stack
  PopResult dblNumber1
  
  ' Return the result of the evaluation
  Evaluate = dblNumber1

PROC_EXIT:
  Exit Function

PROC_ERR:
  MsgBox "Error: " & Err.Number & ". " & Err.Description, , _
    "Evaluate"
  Resume PROC_EXIT

End Function

Private Function CharType() As tokCharType
  ' Comments  : This function determines the type of the current character
  ' Parameters: None
  ' Returns   : The type of the current character
  ' Source    : Total VB SourceBook 6
  '
  On Error GoTo PROC_ERR

  If InStr(mstrSeparatorChars, mstrCurrentChar) > 0 Then
    CharType = tokSeparatorChar
  ElseIf InStr(mstrWhiteSpaceChars, mstrCurrentChar) > 0 Then
    CharType = tokWhiteSpaceChar
  Else
    CharType = tokTokenChar
  End If

PROC_EXIT:
  Exit Function

PROC_ERR:
  MsgBox "Error: " & Err.Number & ". " & Err.Description, , _
    "CharType"
  Resume PROC_EXIT

End Function

Private Sub ConvertToPostfix()
  ' Comments  : This procedure converts the expression to postfix and stores
  '             the result in m_strPostfix
  ' Parameters: None
  ' Returns   : Nothing
  ' Source    : Total VB SourceBook 6
  '
  Dim strToken As String
  Dim strSeparatorUsed As String
  Dim strNewExpression As String
  
  On Error GoTo PROC_ERR

  ' Set the characters to use as separators
  mstrSeparatorChars = "*/+-%^()"
  ' Set the characters to treat as whitespace
  mstrWhiteSpaceChars = " "
  
  ' Set the string used by the tokenizer
  SetTokenString m_strExpression
    
  ' Pad the expression with spaces
  Do While GetNextToken(strToken, strSeparatorUsed)

    ' If strToken is not a null string, assign it to strNewExpression, padded
    ' with a space
    If strToken <> vbNullString Then
      strNewExpression = strNewExpression & strToken & " "
    End If
    
    ' If strSeparator is not a null string, assign it to strNewExpression,
    ' padded with a space
    If strSeparatorUsed <> vbNullString Then
      strNewExpression = strNewExpression & strSeparatorUsed & " "
    End If
    
  Loop
  
  ' Reset the characters used as separators. Because the string is now padded
  ' with spaces, we can ignore the separator characters
  mstrSeparatorChars = ""
  
  ' Set the string used by the tokenizer
  SetTokenString strNewExpression
  
  ' Retrieve the first token in the string
  GetNextToken mstrToken, mstrSeparatorUsed
  
  ' Parse the expression
  ParseExpression

PROC_EXIT:
  Exit Sub

PROC_ERR:
  MsgBox "Error: " & Err.Number & ". " & Err.Description, , _
    "ConvertToPostfix"
  Resume PROC_EXIT

End Sub

Private Function GetNextToken( _
  strToken As String, _
  strSeparatorUsed As String) _
  As Boolean
  ' Comments  : This function returns the next token in the string identified
  '             by the text property
  ' Parameters: strToken - The token retrieved from the string
  '             strSeparatorUsed - The separator used to break the word
  ' Returns   : True if there are more tokens, False otherwise
  ' Source    : Total VB SourceBook 6
  '

  Dim fContinue As Boolean
  
  On Error GoTo PROC_ERR
  
  ' reset the internal token string
  mstrInternalToken = vbNullString
  ' initialize to null
  strSeparatorUsed = vbNullString
  
  ' initialize state
  meState = tokWhiteSpace
  ' initialize previous m_strQuoteChars char

  fContinue = True
  GetNextToken = True
  
  ' Check to see if we are passed the end of the string
  If mlngPos > mlngLen Then
    GetNextToken = False
    fContinue = False
  End If
  
  ' While we have not passed the end of string
  Do While mlngPos <= mlngLen And fContinue = True
    ' Get the current character
    mstrCurrentChar = Mid$(mstrTokenText, mlngPos, 1)

    ' Handle the current type of character
    Select Case CharType

      Case tokSeparatorChar
        fContinue = HandleSeparator(strSeparatorUsed)
        
      Case tokWhiteSpaceChar
        fContinue = HandleWhiteSpace
        
      Case Else
        fContinue = HandleToken
        
    End Select
  Loop
  ' return the token in the strToken parameter
  strToken = mstrInternalToken
  
PROC_EXIT:
  Exit Function

PROC_ERR:
  MsgBox "Error: " & Err.Number & ". " & Err.Description, , _
    "GetNextToken"
  Resume PROC_EXIT

End Function

Private Function HandleSeparator(strSeparatorUsed As String) As Boolean
  ' Comments  : This function handles a separator character
  ' Parameters: strSeparatorUsed - The separator encountered
  ' Returns   : True if parsing should continue, or False if it should stop
  ' Source    : Total VB SourceBook 6
  '
  On Error GoTo PROC_ERR

  ' Increment the current position
  mlngPos = mlngPos + 1
  ' return the separator encountered
  strSeparatorUsed = mstrCurrentChar
  ' stop parsing the token
  HandleSeparator = False

PROC_EXIT:
  Exit Function

PROC_ERR:
  MsgBox "Error: " & Err.Number & ". " & Err.Description, , _
    "HandleSeparator"
  Resume PROC_EXIT

End Function

Private Function HandleToken() As Boolean
  ' Comments  : This function handles a token character
  ' Parameters: None
  ' Returns   : Nothing
  ' Source    : Total VB SourceBook 6
  '
  On Error GoTo PROC_ERR
  
  ' continue parsing by default
  HandleToken = True
  
  Select Case (meState)
    Case tokWhiteSpace
      ' switch to the token state
      meState = tokToken
      ' Append the current character to the internal token string
      mstrInternalToken = mstrInternalToken & LCase(mstrCurrentChar)
      ' Increment the current position
      mlngPos = mlngPos + 1
      
    Case tokToken
      ' Append the current character to the internal token string
      mstrInternalToken = mstrInternalToken & LCase(mstrCurrentChar)
      ' Increment the current position
      mlngPos = mlngPos + 1

    Case tokUndefined
      ' Finish parsing this token
      HandleToken = False
  End Select

PROC_EXIT:
  Exit Function

PROC_ERR:
  MsgBox "Error: " & Err.Number & ". " & Err.Description, , _
    "HandleToken"
  Resume PROC_EXIT

End Function

Private Function HandleWhiteSpace() As Boolean
  ' Comments  : This function handles a whitespace character
  ' Parameters: None
  ' Returns   : True if parsing should continue, or False if it should stop
  ' Source    : Total VB SourceBook 6
  '
  On Error GoTo PROC_ERR

  HandleWhiteSpace = True
  
  Select Case (meState)
    Case tokWhiteSpace, tokUndefined
      ' Increment the current position
      mlngPos = mlngPos + 1
      
    Case tokToken
      meState = tokUndefined
      ' Increment the current position
      mlngPos = mlngPos + 1
      
  End Select
  
PROC_EXIT:
  Exit Function

PROC_ERR:
  MsgBox "Error: " & Err.Number & ". " & Err.Description, , _
    "HandleWhiteSpace"
  Resume PROC_EXIT
  
End Function

Private Sub ParseExpression()
  ' Comments  : This procedure parses an expression
  ' Parameters: None
  ' Returns   : Nothing
  ' Source    : Total VB SourceBook 6
  '
  Dim strToken As String
    
  On Error GoTo PROC_ERR
  
  ' Parse the term first
  ParseTerm
  
  ' while the current token is a + or -
  Do While mstrToken = mcstrAdd Or mstrToken = mcstrSub
    ' save the current token
    strToken = mstrToken
    ' Get the next token
    GetNextToken mstrToken, mstrSeparatorUsed
    ' Parse the term
    ParseTerm
    'Append the saved token to the string
    m_strPostfix = m_strPostfix & strToken & " "
  Loop
  
PROC_EXIT:
  Exit Sub

PROC_ERR:
  MsgBox "Error: " & Err.Number & ". " & Err.Description, , _
    "ParseExpression"
  Resume PROC_EXIT
  
End Sub

Private Sub ParseFactor()
  ' Comments  : This procedure parses a factor
  ' Parameters: None
  ' Returns   : Nothing
  ' Source    : Total VB SourceBook 6
  '
  Dim strToken As String
  Dim fUnaryMinus As Boolean
    
  On Error GoTo PROC_ERR
    
  ' Check for unary + or -
  If mstrToken = mcstrUnaryMinus Or mstrToken = mcstrUnaryPlus Then
      
    ' if this is a unary minus, remember it
    If mstrToken = mcstrUnaryMinus Then
      fUnaryMinus = True
    End If
    
    ' if we found a unary + or -, get the next token
    GetNextToken mstrToken, mstrSeparatorUsed
  End If
  
  ' If the paren minus flag is set, flip the sign of the unary minus
  If mfNegateParenOps Then
    fUnaryMinus = Not fUnaryMinus
  End If
  
  ' If we have encountered a left parent
  If mstrToken = mcstrLParen Then
    ' If we encountered a unary minus just in front of the parenthesis, then
    ' save that fact in a module level variable
    If fUnaryMinus Then
      mfNegateParenOps = True
    End If
    
    GetNextToken mstrToken, mstrSeparatorUsed
    
    ' Recursively parse the expression
    ParseExpression
    
    ' If the current token is right paren, the expression is correct
    ' otherwise the behavior is undefined
    If mstrToken = mcstrRParen Then
      ' get the next token
      GetNextToken mstrToken, mstrSeparatorUsed
      ' reset the unary minus on parenthesis flag
      mfNegateParenOps = False
    End If
  
  Else
    ' If the current token is a function
    If mstrToken = mcstrSin Or mstrToken = mcstrCos Or mstrToken = mcstrTan Or _
      mstrToken = mcstrAtan Or mstrToken = mcstrExp Or mstrToken = mcstrLog Then
      ' save the current token
      strToken = mstrToken
      ' get the next token
      GetNextToken mstrToken, mstrSeparatorUsed
      ' Get the factors inside of the parenthesis
      ParseFactor
      
      ' If a unary minus was encountered, include it in the function
      If fUnaryMinus Then
        m_strPostfix = m_strPostfix & mcstrUnaryMinus & strToken & " "
      Else
        m_strPostfix = m_strPostfix & strToken & " "
      End If
    
    Else
      ' The current token is a number
      ' If a unary minus was encountered, include it in the number
      If fUnaryMinus Then
        m_strPostfix = m_strPostfix & mcstrUnaryMinus & mstrToken & " "
      Else
        m_strPostfix = m_strPostfix & mstrToken & " "
      End If

      ' get the next token
      GetNextToken mstrToken, mstrSeparatorUsed
    End If
  End If

PROC_EXIT:
  Exit Sub

PROC_ERR:
  MsgBox "Error: " & Err.Number & ". " & Err.Description, , _
    "ParseFactor"
  Resume PROC_EXIT

End Sub

Private Sub ParseTerm()
  ' Comments  : This procedure parses a term. A term is something like 2+7,
  '             or 1 + sin(1), while a token is the smallest meaningful piece
  '             of data, such as 2 or sin
  ' Parameters: None
  ' Returns   : Nothing
  ' Source    : Total VB SourceBook 6
  '
  Dim strToken As String
  
  On Error GoTo PROC_ERR
  
  'Parse the factor first
  ParseFactor
  
  ' while the current token is a * or / or % or ^
  Do While mstrToken = mcstrMult Or mstrToken = mcstrDiv Or mstrToken = mcstrMod _
    Or mstrToken = mcstrPower
    ' save the current token
    strToken = mstrToken
    ' get the next token
    GetNextToken mstrToken, mstrSeparatorUsed
    ' parse the factor
    ParseFactor
    ' Append the saved token to the string
    m_strPostfix = m_strPostfix & strToken & " "
  Loop

PROC_EXIT:
  Exit Sub

PROC_ERR:
  MsgBox "Error: " & Err.Number & ". " & Err.Description, , _
    "ParseTerm"
  Resume PROC_EXIT

End Sub

Private Function PopResult(varData As Double) As Boolean
  ' Comments  : Pops an item off the stack
  ' Parameters: varData - returns the data contained on the stack
  ' Returns   : True if an item was popped off the stack, or False if there is
  '             no data left
  ' Source    : Total VB SourceBook 6
  '
  On Error GoTo PROC_ERR
  
  ' If there is data remaining on the stack
  If mintTop > 0 Then
    
    ' decrement the index into the array by 1
    mintTop = mintTop - 1
    
    varData = mavarData(mintTop)

    PopResult = True
  Else
    'There is no data on the stack
    'Return False
    PopResult = False
    'Set the data parameter to zero
    varData = 0
  End If

PROC_EXIT:
  Exit Function

PROC_ERR:
  MsgBox "Error: " & Err.Number & ". " & Err.Description, , _
    "PopResult"
  Resume PROC_EXIT
  
End Function

Private Sub PushResult(varData As Double)
  ' Comments  : Pushes an item onto the stack
  ' Parameters: varData - the data to push onto the stack
  ' Returns   : Nothing
  ' Source    : Total VB SourceBook 6
  '
  On Error GoTo PROC_ERR

  mavarData(mintTop) = varData
  
  ' increment the index into the array by 1
  mintTop = mintTop + 1

PROC_EXIT:
  Exit Sub

PROC_ERR:
  MsgBox "Error: " & Err.Number & ". " & Err.Description, , _
    "PushResult"
  Resume PROC_EXIT

End Sub

Private Sub SetTokenString(strTokenText As String)
  ' Comments  : Set the string used by the tokenizer
  ' Parameters: strTokenText - The string to tokenize
  ' Returns   : Nothing
  ' Source    : Total VB SourceBook 6
  '
  On Error GoTo PROC_ERR

  mstrTokenText = strTokenText
  
  ' Reset internal tokenizer variables
  mlngPos = 1
  mlngLen = Len(strTokenText)

PROC_EXIT:
  Exit Sub

PROC_ERR:
  MsgBox "Error: " & Err.Number & ". " & Err.Description, , _
    "SetTokenString"
  Resume PROC_EXIT

End Sub

